from netCDF4 import Dataset
import datetime as dt
import unittest

import numpy as np

from cis.cis_main import aggregate_cmd
from cis.test.integration.base_integration_test import BaseIntegrationTest
from cis.test.integration_test_data import *
from cis.test.utils_for_testing import *
from cis.parse import parse_args


class BaseAggregationTest(BaseIntegrationTest):
    def check_grid_aggregation(self, lat_start, lat_end, lat_delta, lon_start, lon_end, lon_delta,
                               lat_name='latitude', lon_name='longitude'):
        self.ds = Dataset(self.OUTPUT_FILENAME)
        expected_lat_bnds = np.array([[y, y + lat_delta] for y in np.arange(lat_start, lat_end, lat_delta)])
        expected_lon_bnds = np.array([[x, x + lon_delta] for x in np.arange(lon_start, lon_end, lon_delta)])
        lat_bnds = self.ds.variables[lat_name + '_bnds']
        lon_bnds = self.ds.variables[lon_name + '_bnds']
        try:
            assert_that(np.allclose(lat_bnds[:], expected_lat_bnds))
        except AssertionError:
            # For some reason we occasionally need to reverse the order...
            expected_lat_bnds = np.array(([[y + lat_delta, y] for y in np.arange(lat_start, lat_end, lat_delta)]))[::-1]
            assert_that(np.allclose(lat_bnds[:], expected_lat_bnds))
        try:
            assert_that(np.allclose(lon_bnds[:], expected_lon_bnds))
        except AssertionError:
            expected_lon_bnds = np.array([[x + lon_delta, x] for x in np.arange(lon_start, lon_end, lon_delta)])[::-1]
            assert_that(np.allclose(lon_bnds[:], expected_lon_bnds))
        self.ds.close()

    def check_temporal_aggregation(self, time_start, time_end, time_delta, time_name='time'):
        self.ds = Dataset(self.OUTPUT_FILENAME)
        # Convert from time to days after..

        def convert_to_days_since_cis_epoch(date_time):
            cis_standard = dt.datetime(1600, 1, 1, 0, 0, 0)
            return (date_time - cis_standard).total_seconds() / dt.timedelta(days=1).total_seconds()

        time_start = convert_to_days_since_cis_epoch(time_start)
        time_end = convert_to_days_since_cis_epoch(time_end)
        time_delta = time_delta.total_seconds() / dt.timedelta(days=1).total_seconds()

        expected_time_bnds = np.array([[t, t + time_delta] for t in np.arange(time_start, time_end, time_delta)])
        time_bnds = self.ds.variables[time_name + '_bnds']
        assert_that(time_bnds.shape == expected_time_bnds.shape)
        assert_that(np.allclose(time_bnds[:], expected_time_bnds))
        self.ds.close()

    def do_spatial_aggregate(self, variable, filename, lat_start, lat_end, lat_delta, lon_start, lon_end, lon_delta):
        grid = 'x=[%s,%s,%s],y=[%s,%s,%s]' % (lon_start, lon_end, lon_delta, lat_start, lat_end, lat_delta)
        arguments = ['aggregate', variable + ':' + escape_colons(filename) + ':kernel=mean', grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)

    def do_temporal_aggregate(self, variable, filename, t_start, t_end, str_delta):
        grid = 't=[%s,%s,%s]' % (t_start.isoformat(), t_end.isoformat(), str_delta)
        arguments = ['aggregate', variable + ':' + escape_colons(filename) + ':kernel=mean', grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)


class TestAggregation(BaseAggregationTest):
    """
    This class provides integration testing for the Aggregation command, ensuring that for one specific product
    (Aerosol CCI), the aggregation is correct and the results are what we expect.
    """

    def test_aggregation_over_latlon(self):
        import numpy.ma as ma
        # Takes 20s
        variable = 'AOD550,AOD870'
        filename = valid_aerosol_cci_filename
        lat_start, lat_end, lat_delta = -40, 80, 10  # 12
        lon_start, lon_end, lon_delta = -20, 140, 10  # 16
        self.do_spatial_aggregate(variable, filename, lat_start, lat_end, lat_delta, lon_start, lon_end, lon_delta)
        self.check_grid_aggregation(lat_start, lat_end, lat_delta, lon_start, lon_end, lon_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

        # Test some specific values: these are previous outputs from this aggregate command which have been compared
        # against the original data and found to 'look about right' and therefore used as a baseline. If these values
        # change we should consider being concerned.
        i = float('inf')
        arr_550 = np.array([[[0.1096153093624289, i, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.405634596128, i, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.175624571614, 0.152486790551, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.09970282124016773, 0.1629985723514454, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.19553465147813162, 0.405634596128, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.7052307901349156, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.5859629385860146, 0.4436802918665714, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.29588747474573723, 0.2914036842614198, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.13670555069561927, 0.1756245641523532, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, 0.1524867860807313, 0.15664356061755572, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, i, 0.1415671800312243, 0.8644281417604477, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, i, i, 0.18689890454212824, i, i, i, i, i, i, i, i, i, i, 0.17285842509710625]]])
        arr_870 = np.array([[[0.0869540874214068, i, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.239833246086, i, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.106763551248, 0.0872576713562, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.0626557093240659, 0.0976959701925822, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [0.110700003802776, 0.239833255113166, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.639897634027878, i, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.533210717307752, 0.386442222587703, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.226935877285485, 0.228134018585045, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, 0.0762777089351608, 0.106399147962141, i, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, 0.087257672391004, 0.0883575717452914, i, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, i, 0.189694136753678, 0.680416197050363, i, i, i, i, i, i, i, i, i, i, i],
                             [i, i, i, i, 0.127948367761241, i, i, i, i, i, i, i, i, i, i, 0.0958142693422429]]])
        arr_550 = ma.masked_invalid(arr_550)
        arr_870 = ma.masked_invalid(arr_870)
        self.ds = Dataset(self.OUTPUT_FILENAME)
        data_550 = self.ds.variables['AOD550']
        assert_arrays_almost_equal(data_550[:], arr_550.reshape((16,12,1)))
        data_870 = self.ds.variables['AOD870']
        assert_arrays_almost_equal(data_870[:], arr_870.reshape((16,12,1)))

    def test_aggregation_over_time(self):
        import numpy.ma as ma
        # Takes 14s
        variable = 'AOD550,AOD870'
        filename = valid_aerosol_cci_filename
        t_start, t_end = dt.datetime(2008, 6, 12, 10, 15, 0), dt.datetime(2008, 6, 12, 10, 35, 0)
        t_delta = dt.timedelta(minutes=1)
        str_delta = 'PT1M'
        self.do_temporal_aggregate(variable, filename, t_start, t_end, str_delta)
        self.check_temporal_aggregation(t_start, t_end, t_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

        # Test some specific values: these are previous outputs from this aggregate command which have been compared
        # against the original data and found to 'look about right' and therefore used as a baseline. If these values
        # change we should consider being concerned.
        i = float('inf')
        arr_550 = np.array( [[[0.2724732138893821, 0.2845287322998047, 0.18888297843565602,
                               0.1977633680318641, 0.1528509303068309, 0.21690222715974708,
                               0.3218141342953317, 0.3296216605627661, 0.42861915269278755,
                               0.6922391289425149, 1.1032210622605763, 0.7501978423274019,
                               0.4114444071076453, i, 0.5656480511415352, 0.3725444489875726,
                               0.20023197438583437, 0.12606849266025288, 0.10038993358612061,
                               0.08870828946431478]]])
        arr_870 = np.array([[[0.15527870871803978, 0.177825194138747, 0.11566017769478462,
                               0.12211483395949896, 0.09108650039336486, 0.139907898276485,
                               0.24998173519681813, 0.29386406702711093, 0.38747846944250536,
                               0.6303785410104443, 0.999659818761489, 0.6851485338292652,
                               0.37053351032777904, 0.2074383, 0.336047126251517, 0.22054079781591365,
                               0.11859528582197681, 0.07632432809566862, 0.062386703491210935,
                               0.053831556108262804]]])
        arr_550 = ma.masked_invalid(arr_550)
        self.ds = Dataset(self.OUTPUT_FILENAME)
        data_550 = self.ds.variables['AOD550']
        assert_arrays_almost_equal(data_550[:], arr_550, 1.0e-2)
        data_870 = self.ds.variables['AOD870']
        assert_arrays_almost_equal(data_870[:], arr_870, 1.0e-2)

    def test_aggregation_of_multiple_variables_gives_same_result(self):
        # JASCIS-281
        variable = 'AOT_440'
        filename = another_valid_aeronet_filename
        t_start, t_end = dt.datetime(1999, 0o1, 0o1), dt.datetime(1999, 12, 29)
        t_delta = dt.timedelta(hours=6)
        str_delta = 'PT6H'
        self.do_temporal_aggregate(variable, filename, t_start, t_end, str_delta)
        self.check_temporal_aggregation(t_start, t_end, t_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

        # Read the aggregated data
        self.ds = Dataset(self.OUTPUT_FILENAME)
        aot_440 = self.ds.variables['AOT_440'][:]

        # Clean up the old aggregation
        self.ds.close()
        os.remove(self.OUTPUT_FILENAME)

        # Now redo the aggregation, but for all AOT variables
        self.do_temporal_aggregate('AOT*', filename, t_start, t_end, str_delta)

        # Read the new data in
        self.ds = Dataset(self.OUTPUT_FILENAME)
        aot_440_multi_var = self.ds.variables['AOT_440'][:]

        # And check they are the same...
        assert_arrays_almost_equal(aot_440, aot_440_multi_var)


class TestSpatialAggregationByDataProduct(BaseAggregationTest):
    """
    These tests are necessarily limited in scope - we can't check the output is correct for all of them (this is done
    for one specific product in the above class 'TestAggregation') so the best we can reasonably do is:
    - Check the aggregation completes with no errors.
    - Check that the output coordinates reflect the correct aggregation bins.
    - Check that all variables requested are in the output file.
    """

    def test_aggregate_Aerosol_CCI_with_lon_wraparound(self):
        variable = valid_aerosol_cci_variable
        filename = valid_aerosol_cci_filename
        lon_min, lon_max, lon_delta = 10, 350, 10
        lat_min, lat_max, lat_delta = -90, 90, 30
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_Cloud_CCI_with_dimension_vars(self):
        variable = 'time,solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        lon_min, lon_max, lon_delta = 84, 99, 1
        lat_min, lat_max, lat_delta = -6, 6, 1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, ['aggregated_time', 'solar_zenith_view_no1'])

    def test_aggregate_Cloud_CCI(self):
        variable = 'satellite_zenith_view_no1,solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        lon_min, lon_max, lon_delta = 84, 99, 1
        lat_min, lat_max, lat_delta = -6, 6, 1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_Cloud_CCI_for_comparison_with_collocation(self):
        """
        Takes ~80s on aopposxlap18. This test is primarily to compare the speed of ungridded->gridded collocation and
        aggregation, which should in principle be just as quick as each other. This test mirrors the
        test_cloud_cci_onto_NetCDF_Gridded test in test_colocate.py.
        """
        variables = ','.join([valid_cloud_cci_variable, valid_cloud_cci_8_bit_variable])
        filename = valid_cloud_cci_filename
        grid = 'x=[0, 358.125, 1.875],y=[-90, 90, 1.25],t=[2007-06-01,2007-06-30,PT3H]'
        arguments = ['aggregate', variables + ':' + escape_colons(filename) + ':kernel=mean', grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)

        # This check can't deal with the combined temporal and spatial aggregation so skip it
        #self.check_grid_aggregation(0, 358.125, 1.875, -90, 90, 1.25)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variables.split(','))

    def test_aggregate_NCAR_RAF(self):
        # Takes 513s
        variable = valid_NCAR_NetCDF_RAF_variable
        filename = valid_NCAR_NetCDF_RAF_filename
        lon_min, lon_max, lon_delta = -160, -157, 1
        lat_min, lat_max, lat_delta = 30, 50, 5
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    def test_aggregate_GASSP(self):
        variable = '*'
        filename = valid_GASSP_aeroplane_filename
        lon_min, lon_max, lon_delta = -94, 95, 0.1
        lat_min, lat_max, lat_delta = 30, 31, 0.1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    def test_aggregate_GASSP2(self):
        # see issue JASCIS-144
        variable = valid_GASSP_not_entirely_correct_variable
        filename = valid_GASSP_not_entirely_correct_filename
        lon_min, lon_max, lon_delta = -180, 180, 10
        lat_min, lat_max, lat_delta = -90, 90, 10
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    def test_aggregate_Aeronet(self):
        # Takes 50s
        variable = 'AOT_*'
        filename = valid_aeronet_filename
        lon_min, lon_max, lon_delta = -1.5, 1.4, 0.01
        lat_min, lat_max, lat_delta = 15, 15.5, 0.1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    def test_aggregate_netCDF_gridded_HadGem(self):
        # Test aggregating a gridded file, it should work but throw a deprecation
        # Takes 1s
        variable = '*'
        filename = valid_hadgem_filename
        arguments = ['aggregate', variable + ':' + escape_colons(filename) + ':kernel=mean', 'x,y', '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)

        self.check_output_contains_variables(self.OUTPUT_FILENAME, ['od550aer'])

    def test_aggregate_cis_ungridded(self):
        # Takes 1s
        variable = '*'
        filename = valid_cis_ungridded_output_filename
        lon_min, lon_max, lon_delta = 1, 3, 0.3
        lat_min, lat_max, lat_delta = 41, 42, 0.1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    @skip_pyhdf
    def test_aggregate_MODIS_L2(self):
        # Takes 66s
        variable = '*'
        filename = valid_modis_l2_filename
        lon_min, lon_max, lon_delta = -150, 150, 25
        lat_min, lat_max, lat_delta = -72, -63, 1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    @skip_pyhdf
    def test_aggregate_CloudSatPRECIP(self):
        # Takes 190s
        variable = '*'
        filename = valid_cloudsat_PRECIP_file
        lon_min, lon_max, lon_delta = -10, 10, 2
        lat_min, lat_max, lat_delta = 40, 60, 2
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    @unittest.skip("Very resource intensive")
    def test_aggregate_Caliop_L1(self):
        variable = '*'  # This takes over 4 hrs (but does work)
        variable = 'Perpendicular_Attenuated_Backscatter_532'
        filename = valid_caliop_l1_filename
        lon_min, lon_max, lon_delta = 0, 60, 20
        lat_min, lat_max, lat_delta = -30, 30, 20
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    @skip_pyhdf
    def test_aggregate_CloudSatRVOD(self):
        # Takes 200s
        variable = '*'  # Slow and runs out of memory
        variable = 'RVOD_liq_water_path,RVOD_ice_water_path'
        filename = valid_cloudsat_RVOD_file
        lon_min, lon_max, lon_delta = -10, 10, 2
        lat_min, lat_max, lat_delta = 40, 60, 2
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_aggregate_Caliop_L2(self):
        # Takes 37 mins
        variable = '*'
        filename = valid_caliop_l2_filename
        lon_min, lon_max, lon_delta = 0, 60, 6
        lat_min, lat_max, lat_delta = -30, 30, 6
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)

    def test_subset_ASCII(self):
        # Takes 1s
        variable = '*'
        filename = valid_ascii_filename
        lon_min, lon_max, lon_delta = -10, 10, 2
        lat_min, lat_max, lat_delta = 1, 6, 1
        self.do_spatial_aggregate(variable, filename, lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta)


class TestTemporalAggregationByDataProduct(BaseAggregationTest):
    """
    These tests are necessarily limited in scope - we can't check the output is correct for all of them (this is done
    for one specific product in the above class 'TestAggregation') so the best we can reasonably do is:
    - Check the aggregation completes with no errors.
    - Check that the output coordinates reflect the correct aggregation bins.
    - Check that all variables requested are in the output file.
    """

    def test_aggregate_Aerosol_CCI(self):
        # This is tested in the TestAggregation class above
        pass

    def test_aggregate_Cloud_CCI_with_dimension_vars(self):
        # Takes 826s
        variable = 'time,lon,lat,satellite_zenith_view_no1,solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        time_min, time_max = dt.datetime(2008, 7, 1, ), dt.datetime(2008, 8, 1)

        time_delta = dt.timedelta(days=5)
        str_delta = 'P5D'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_Cloud_CCI(self):
        variable = 'solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        time_min, time_max = dt.datetime(2008, 7, 1, ), dt.datetime(2008, 8, 1)

        time_delta = dt.timedelta(days=5)
        str_delta = 'P5D'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_NCAR_RAF(self):
        # Takes 28s
        variable = "LATC,LONC,GGALTC,Time,PSXC,WSC,ATX,ATHR2,CONCD_LWI"
        filename = valid_NCAR_NetCDF_RAF_filename
        time_min, time_max = dt.datetime(2009, 1, 14, 20, 15), dt.datetime(2009, 1, 15, 2, 45)

        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_NCAR_RAF_with_named_time_variable_standard_name(self):
        # Takes 28s
        variable = "LATC,LONC,GGALTC,Time,PSXC,WSC,ATX,ATHR2,CONCD_LWI"
        filename = valid_NCAR_NetCDF_RAF_filename
        time_min, time_max = dt.datetime(2009, 1, 14, 20, 15), dt.datetime(2009, 1, 15, 2, 45)

        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'

        grid = 'time=[%s,%s,%s]' % (time_min.isoformat(), time_max.isoformat(), str_delta)
        arguments = ['aggregate', variable + ':' + escape_colons(filename) + ':kernel=mean', grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)

        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_GASSP(self):
        variable = ",".join(valid_GASSP_aeroplane_vars)
        filename = valid_GASSP_aeroplane_filename
        time_min, time_max = dt.datetime(2006, 9, 27, 20, 15), dt.datetime(2006, 9, 27, 22, 45)
        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_GASSP_aux_coord(self):
        # Takes 1.3s
        filename = cis_test_files['GASSP_aux_coord'].master_filename
        variable = cis_test_files['GASSP_aux_coord'].data_variable_name
        time_min, time_max = dt.datetime(2006, 9, 27, 20, 15), dt.datetime(2006, 9, 27, 22, 45)
        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_Aeronet(self):
        # Takes 2s
        variable = 'AOT_440'
        filename = valid_aeronet_filename
        time_min, time_max = dt.datetime(1999, 1, 1, 0, 0, 0), dt.datetime(1999, 12, 29, 0, 0, 0)
        time_delta = dt.timedelta(hours=6)
        str_delta = 'PT6H'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_Aeronet_multi_var(self):
        # Takes 2s
        variable = 'AOT_1640,AOT_1020,AOT_870,AOT_675,AOT_667,AOT_555,AOT_551,' \
                   'AOT_532,AOT_531,AOT_500,AOT_490,AOT_443,AOT_440,AOT_412,AOT_380,AOT_340'
        filename = valid_aeronet_filename
        time_min, time_max = dt.datetime(1999, 1, 1), dt.datetime(1999, 12, 29)
        time_delta = dt.timedelta(hours=6)
        str_delta = 'PT6H'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))
        self.check_output_vars_are_different(self.OUTPUT_FILENAME, variable.split(',')[:5])

    def test_aggregate_cis_ungridded(self):
        # Takes 2s
        variable = 'AOD550,AOD870,latitude,time'
        filename = valid_cis_ungridded_output_filename
        time_min, time_max = dt.datetime(2008, 6, 12, 10, 18, 37), dt.datetime(2008, 6, 12, 10, 19, 47)
        time_delta = dt.timedelta(seconds=1)
        str_delta = 'PT1S'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_aggregate_MODIS_L2(self):
        # Takes 20 mins
        variable = 'Solar_Zenith,Latitude,Sensor_Azimuth,Longitude'
        filename = valid_modis_l2_filename
        time_min, time_max = dt.datetime(2010, 1, 1, 22, 55, 19), dt.datetime(2010, 1, 1, 22, 58, 44)
        time_delta = dt.timedelta(minutes=1)
        str_delta = 'PT1M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_aggregate_CloudSatPRECIP_with_dimension_variables(self):
        # Takes 28s
        # RuntimeError: NetCDF: String match to name in use
        variable = 'Profile_time,Latitude,Longitude,DEM_elevation,Data_quality'
        filename = valid_cloudsat_PRECIP_file
        time_min, time_max = dt.datetime(2008, 2, 14, 0, 57, 36), dt.datetime(2008, 2, 14, 2, 9, 36)
        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(',') +
                                             ['aggregated_Profile_time'])

    @skip_pyhdf
    def test_aggregate_CloudSatPRECIP(self):
        variable = valid_cloudsat_PRECIP_variable
        filename = valid_cloudsat_PRECIP_file
        time_min, time_max = dt.datetime(2008, 2, 14, 0, 57, 36), dt.datetime(2008, 2, 14, 2, 9, 36)
        time_delta = dt.timedelta(minutes=30)
        str_delta = 'PT30M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [valid_cloudsat_PRECIP_variable])

    @skip_pyhdf
    def test_aggregate_CloudSatRVOD(self):
        # Takes 41s
        variable = 'RVOD_liq_water_path,RVOD_ice_water_path'
        filename = valid_cloudsat_RVOD_file
        time_min, time_max = dt.datetime(2007, 6, 29, 13, 12, 0), dt.datetime(2007, 6, 29, 14, 29, 0)
        time_delta = dt.timedelta(minutes=23)
        str_delta = 'PT23M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @unittest.skip("Very resource intensive")
    def test_aggregate_Caliop_L1(self):
        # This takes a long time to run (we've never seen it complete, even on sci-1).
        variable = 'Perpendicular_Attenuated_Backscatter_532'
        filename = valid_caliop_l1_filename
        time_min, time_max = dt.datetime(2009, 12, 31, 23, 40, 0), dt.datetime(2010, 1, 1, 0, 17, 17)
        time_delta = dt.timedelta(minutes=15)
        str_delta = 'PT15M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_aggregate_Caliop_L2(self):
        # Takes 20s
        variable = 'Perpendicular_Backscatter_Coefficient_532,' \
                   'Perpendicular_Backscatter_Coefficient_Uncertainty_532,Pressure'
        filename = valid_caliop_l2_filename
        time_min, time_max = dt.datetime(2009, 12, 31, 23, 40, 0), dt.datetime(2010, 1, 1, 0, 17, 17)
        time_delta = dt.timedelta(minutes=15)
        str_delta = 'PT15M'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_aggregate_ASCII(self):
        # Takes 1s
        variable = 'value'
        filename = valid_ascii_filename
        time_min, time_max = dt.datetime(2012, 8, 23, 15, 0, 0), dt.datetime(2012, 8, 28)
        time_delta = dt.timedelta(days=1)
        str_delta = 'P1D'
        self.do_temporal_aggregate(variable, filename, time_min, time_max, str_delta)
        self.check_temporal_aggregation(time_min, time_max, time_delta)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))


class TestMomentsKernel(BaseAggregationTest):
    def test_GIVEN_no_kernel_specified_WHEN_aggregate_THEN_moments_used_as_default(self):
        # Takes 1s
        variable = 'AOD550'
        filename = valid_cis_ungridded_output_filename
        lon_min, lon_max, lon_delta = 1, 3, 0.3
        lat_min, lat_max, lat_delta = 41, 42, 0.1
        grid = 'x=[%s,%s,%s],y=[%s,%s,%s]' % (lon_min, lon_max, lon_delta, lat_min, lat_max, lat_delta)
        arguments = ['aggregate', variable + ':' + escape_colons(filename), grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta,
                                    lat_name='latitude', lon_name='longitude')
        expected_vars = ['AOD550', 'AOD550_std_dev', 'AOD550_num_points']
        self.check_output_contains_variables(self.OUTPUT_FILENAME, expected_vars)

    def test_moments_kernel_aggregate_cis_ungridded(self):
        # Takes 1s
        variable = '*'
        filename = valid_cis_ungridded_output_filename
        lon_min, lon_max, lon_delta = 1, 3, 0.3
        lat_min, lat_max, lat_delta = 41, 42, 0.1
        grid = 'x=[%s,%s,%s],y=[%s,%s,%s]' % (lon_min, lon_max, lon_delta, lat_min, lat_max, lat_delta)
        arguments = ['aggregate', variable + ':' + escape_colons(filename) + ':kernel=moments', grid, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        aggregate_cmd(main_arguments)
        self.check_grid_aggregation(lat_min, lat_max, lat_delta, lon_min, lon_max, lon_delta,
                                    lat_name='latitude', lon_name='longitude')
        expected_vars = ['AOD870', 'AOD870_std_dev', 'AOD870_num_points',
                         'AOD550', 'AOD550_std_dev', 'AOD550_num_points']
        self.check_output_contains_variables(self.OUTPUT_FILENAME, expected_vars)



import unittest

import numpy

from cis.data_io.ungridded_data import UngriddedDataList
import cis.test.util.mock as mock
from cis.collocation.col_implementations import DummyCollocator


class TestDummyCollocator(unittest.TestCase):
    
    def test_single_data(self):
        sample = mock.make_regular_2d_ungridded_data()
        data = mock.make_regular_2d_ungridded_data(data_offset=10)
        col = DummyCollocator()
        con = None
        kernel = None

        output = col.collocate(sample, data, con, kernel)

        assert len(output) == 1
        assert numpy.array_equal(output[0].data, data.data)

    def test_list_of_data(self):
        sample = mock.make_regular_2d_ungridded_data()
        data = UngriddedDataList([mock.make_regular_2d_ungridded_data(data_offset=5),
                                  mock.make_regular_2d_ungridded_data(data_offset=10)])
        col = DummyCollocator()
        con = None
        kernel = None

        output = col.collocate(sample, data, con, kernel)

        assert len(output) == 2
        assert numpy.array_equal(output[0].data, data[0].data)
        assert numpy.array_equal(output[1].data, data[1].data)

if __name__ == '__main__':
    unittest.main()



"""
Module to do integration tests of plots, checking that the outputs look right.
"""
from cis.cis_main import plot_cmd
from cis.parse import parse_args
from cis.test.integration_test_data import *
from cis.test.integration.base_integration_test import BaseIntegrationTest
from cis.data_io.products.AProduct import ProductPluginException
from cis.exceptions import InvalidDimensionError
from nose.tools import raises
from unittest import skip

import shutil
import logging

import matplotlib.testing.compare as mcompare
import matplotlib.pyplot as plt

# Whether to display matplotlib output to the screen.
_DISPLAY_FIGURES = False

# plt.switch_backend('agg')


class VisualTest(BaseIntegrationTest):

    _DEFAULT_IMAGE_TOLERANCE = 0.2

    def id(self):
        # Just use the module, class and method name - the root can change depending on where it is run from
        return ".".join(super(VisualTest, self).id().split('.')[-3:])

    def setUp(self):
        # Set force overwrite in case working files are still present
        os.environ['CIS_FORCE_OVERWRITE'] = "True"
        # Make sure we have no unclosed plots from previous tests before
        # generating this one.
        if _DISPLAY_FIGURES:
            plt.close('all')

    def tearDown(self):
        # Pop off the environemnt variable
        os.environ.pop('CIS_FORCE_OVERWRITE')
        # If a plotting test bombs out it can leave the current figure
        # in an odd state, so we make sure it's been disposed of.
        if _DISPLAY_FIGURES:
            plt.close('all')

    def check_graphic(self, tol=_DEFAULT_IMAGE_TOLERANCE):
        """Checks the CRC matches for the current matplotlib.pyplot figure, and closes the figure."""

        test_id = self.id()

        figure = plt.gcf()

        try:
            expected_fname = os.path.join(os.path.dirname(__file__),
                                          'reference', 'visual_tests',
                                          test_id + '.png')

            if not os.path.isdir(os.path.dirname(expected_fname)):
                os.makedirs(os.path.dirname(expected_fname))

            #: The path where the images generated by the tests should go.
            image_output_directory = os.path.join(os.path.dirname(__file__),
                                                  'result_image_comparison')
            if not os.access(image_output_directory, os.W_OK):
                if not os.access(os.getcwd(), os.W_OK):
                    raise IOError('Write access to a local disk is required '
                                  'to run image tests.  Run the tests from a '
                                  'current working directory you have write '
                                  'access to to avoid this issue.')
                else:
                    image_output_directory = os.path.join(os.getcwd(), 'result_image_comparison')
            result_fname = os.path.join(image_output_directory, test_id + '.png')

            if not os.path.isdir(os.path.dirname(result_fname)):
                # Handle race-condition where the directories are
                # created sometime between the check above and the
                # creation attempt below.
                try:
                    os.makedirs(os.path.dirname(result_fname))
                except OSError as err:
                    # Don't care about "File exists"
                    if err.errno != 17:
                        raise

            # Output filename if the test output a file itself (using -o)
            output_fname = os.path.join(os.path.dirname(__file__), test_id+'.png')
            # If the test created an output file itself then move that to the results folder, otherwise create an output
            if os.path.exists(output_fname):
                shutil.move(os.path.join(os.path.dirname(__file__), test_id+'.png'), result_fname)
            else:
                figure.savefig(result_fname)

            if not os.path.exists(expected_fname):
                logging.warn('Created image for test %s' % test_id)
                shutil.copy2(result_fname, expected_fname)

            try:
                err = mcompare.compare_images(expected_fname, result_fname, tol=tol)
            except ValueError:
                failed_name = mcompare.make_test_filename(result_fname, 'failed-diff')
                shutil.copy2(os.path.join(os.path.dirname(__file__),
                                          'reference', 'kitten.png'), failed_name)
                err = "Images differ in size and so are not comparable"

            if _DISPLAY_FIGURES:
                if err:
                    print(('Image comparison would have failed. Message: %s' % err))
                plt.show()
            else:
                assert not err, 'Image comparison failed. Message: %s' % err

        finally:
            plt.close()


class TestPlotVisual(VisualTest):

    def test_iris_comparative_scatter(self):
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":color=green,itemstyle=^,itemwidth=40",
                     "snow:" + escape_colons(valid_2d_filename), "--type", "comparativescatter",
                     "--logx", "--logy", "--output", self.id() + ".png"]

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_contour(self):
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":cmap=RdBu", "--type", "contour",
                     "--xlabel", "Overidden X Label", "--title", "Overidded Title",
                    "--height", "5", "--width", "5", "--xmin", "15", "--xmax", "45", "--xstep", "20", "--cbarorient",
                     "vertical", "--fontsize", "10", "--output", self.id() + ".png"]

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_contourf(self):
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename), "--type", "contourf",
                     "--ylabel", "Overidden Y Label", "--height", "5", "--width", "10", "--ymin", "15", "--ymax", "45",
                     "--ystep", "10", "--nocolourbar", "--output", self.id() + ".png"]

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_heatmap(self):
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":cmap=RdBu", "--type", "heatmap",
                     "--ylabel", "Overidden Y", "--title", "OveriddenTitle",
                     "--height", "3.5", "--width", "3.5", "--vmax", "0.000135", "--fontsize", "10", "--output",
                     self.id() + ".png"]

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_heatmap_force_minus_180_to_180(self):
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":cmap=RdBu", "--type", "heatmap",
                     "--ylabel", "Overidden Y", "--title", "OveriddenTitle", "--xmin", "-180", "--xmax", "180",
                     "--height", "3.5", "--width", "3.5", "--vmax", "0.000135", "--fontsize", "10", "--output",
                     self.id() + ".png"]

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_histogram(self):
        opts = "--xbins 5 --ymin 1 --logy --output".split() + [self.id() + ".png"]
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename) + ":color=red,itemstyle=step,label=overridenlabel",
                    "snow:" + escape_colons(valid_1d_filename) + ":color=green,itemstyle=step", "--type", "histogram"] + opts

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_histogram2d(self):
        opts = "--cmap RdBu --ylabel overiddeny --title overiddentitle --xmin 0.000002 --xmax 0.000006 " \
               "--ybins 20 --output ".split() + [self.id() + ".png"]
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename), 
                     "snow:" + escape_colons(valid_1d_filename), "--type", "histogram2d"] + opts

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_many_lines(self):
        opts = "--xlabel overiddenxlabel --ylabel overiddenylabel --width 10 --logy --grid " \
               "--ymin 0.00000000001 --xmax 50 --output".split() + [self.id() + ".png"]
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename) + ":color=red,itemwidth=5",
                     "snow:" + escape_colons(valid_1d_filename) + ":itemstyle=dashed,itemwidth=5,label=overiddenlabel2"] + opts

        main_arguments = parse_args(arguments)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_one_line(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--xlabel overiddenxlabel --title overiddentitle --height 5 --width 10 --logx" \
               " --grid".split()
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename) + ":itemstyle=dashed,label=overiddenlabel,itemwidth=4"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_one_line_with_step(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--ystep 0.00001 --xstep 25".split()
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename) ]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_scatter(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter --title overiddentitle --fontsize 10 --width 7 --ymin=-0.0005" \
               " --ymax 0.0005".split()
        arguments = ["plot", "rain:" + escape_colons(valid_1d_filename) + ":itemstyle=^,itemwidth=30"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_scatter2d(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter2d --fontsize 7 --cbarorient vertical --grid --ymin 0 --ymax 90 --vmin 0" \
               " --cbarorient vertical".split()
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":cmap=RdBu",
                     "snow:" + escape_colons(valid_2d_filename) + ":label=snowlabel,cmap=RdBu"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_comparative_scatter(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = " --type comparativescatter --xlabel overiddenx --ylabel overiddeny --title overiddentitle --fontsize 7" \
               " --height 5 --width 5 --xmin 0 --xmax 1 --ymin 0 --ymax 0.5 --grid".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename) + ":color=red,itemstyle=s,itemwidth=40",
                     "AOT_870:" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_taylor_diagram(self):
        output_file_opt = ["--output", self.id() + ".png"]
        # You can change the fontsize, height, width and title, but the labels and limits are ignored
        opts = " --type taylor --bias color --xlabel overiddenx --title overiddentitle --fontsize 17" \
               " --height 15 --width 15 --xmin 0 --xmax 1 --ymin 0 --ymax 0.5 --grid".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename),
                     "AOT_870:" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_contour(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type contour --ylabel overiddenylabel --title overiddentitle --vmin 0 --cbarorient vertical --grid" \
               " --xaxis Latitude --yaxis Height".split()
        arguments = ["plot", "RVOD_liq_water_content:" + escape_colons(valid_cloudsat_RVOD_file) + ":cmap=RdBu"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_contourf(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type contourf --xlabel overiddenxlabel --fontsize 15 --height 10 --ymin 0 --ymax 10000" \
               " --vmin 0 --cbarorient horizontal --grid --xaxis Latitude --yaxis Height".split()
        arguments = ["plot", "RVOD_liq_water_content:" + escape_colons(valid_cloudsat_RVOD_file) + ":cmap=RdBu,itemwidth=4"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_histogram(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type histogram --xlabel overiddenx --fontsize 10 --height 10 --width 10 --xmin 0 --xmax 1.5" \
               " --xbins 20 --grid".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename) + ":itemstyle=step"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_histogram_bin_width(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = " --type histogram --xbins 3".split()
        arguments = ["plot", 
                     "AOT_440:" + escape_colons(valid_aeronet_filename) + ":itemstyle=step",
                     "AOT_870:" + escape_colons(valid_aeronet_filename) + ":itemstyle=step"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_histogram2d(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type histogram2d --xlabel overridenx --ylabel overiddeny --cbarlabel overiddencbarlabel " \
               "--title overiddentitle --fontsize 7 --width 10 --xmin 0 --xmax 2 --xbins 20 --ymin 0 --ymax 1.5" \
               " --ybins 20 --vmin 60 --vmax 480 --vstep 30 --cbarorient vertical --grid".split()
        arguments = ["plot", 
                     "AOT_440:" + escape_colons(valid_aeronet_filename), 
                     "AOT_870:" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_histogram2d_doesnt_plot_coastlines(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type histogram2d".split()
        arguments = ["plot", 
                     "RVOD_liq_water_content:" + escape_colons(valid_cloudsat_RVOD_file), 
                     "Height:" + escape_colons(valid_cloudsat_RVOD_file)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_many_scatter_points(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter --ylabel overiddenylabel --ymin 0.1 --ymax 1 --logy".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename),
                     "AOT_870:" + escape_colons(valid_aeronet_filename) + ":itemstyle=x,itemwidth=20"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_many_scatter_points_given_color(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter --ylabel overiddenylabel --ymin 0.1 --ymax 1 --logy".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename),
                     "AOT_870:" + escape_colons(valid_aeronet_filename) + ":color=blue,itemstyle=x,itemwidth=4"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_many_lines(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type line --ylabel overiddenylabel --ymin 0.1 --ymax 1 --logy".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename) + ":color=green,itemstyle=dotted,itemwidth=2",
                     "AOT_870:" + escape_colons(valid_aeronet_filename) + ":itemstyle=dashed,itemwidth=2",
                     "AOT_1020:" + escape_colons(valid_aeronet_filename) + ":color=red,itemwidth=2"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_one_line(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type line --xlabel overiddenx --title overiddentitle --fontsize 7 --height 7" \
               " --ymin 0.1 --ymax 1 --logy".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename) + ":itemstyle=dashed,label=overiddenlabel,itemwidth=4"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_scatter(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter --xlabel overiddenx --ylabel overiddeny --title overiddentitle" \
               " --fontsize 15 --height 5 --width 6 --logy --grid".split()
        arguments = ["plot", "AOT_440:" + escape_colons(valid_aeronet_filename) + ":itemwidth=2"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_scatter2d(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type scatter2d --title overiddentitle --vmax 175 --logy --xaxis Latitude" \
               " --yaxis Height".split()
        arguments = ["plot", "RVOD_liq_water_content:" + escape_colons(valid_cloudsat_RVOD_file) + ":cmap=RdBu,itemwidth=20"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_other_longitude_wrapping_0_360(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.plotting.formatted_plot import Plotter

        data = make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=-30, lat_max=30)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90)

        self.check_graphic()

    def test_other_longitude_wrapping_0_360_forced_minus_180_to_180(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.plotting.formatted_plot import Plotter

        data = make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=-30, lat_max=30)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90, xmin=-180, xmax=180)

        self.check_graphic()

    def test_other_longitude_wrapping_minus_180_180(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.plotting.formatted_plot import Plotter
        from cis.test.util.mock import make_regular_2d_ungridded_data

        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90)

        self.check_graphic()

    @skip("Known failure")
    def test_other_longitude_wrapping_minus_180_180_forced_0_to_360(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.plotting.formatted_plot import Plotter
        from cis.test.util.mock import make_regular_2d_ungridded_data

        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90, xmin=0, xmax=360)

        self.check_graphic()

    def test_bluemarble_minus_180_180(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.plotting.formatted_plot import Plotter
        from cis.test.util.mock import make_regular_2d_ungridded_data

        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90, nasabluemarble=True)

        self.check_graphic()

    def test_bluemarble_0_360(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.plotting.formatted_plot import Plotter

        data = make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=-30, lat_max=30)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                layer_opts=[{'itemwidth': 50}], ymin=-90, ymax=90, nasabluemarble=True)

        self.check_graphic()

    def test_coastline_color(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.plotting.formatted_plot import Plotter

        data = make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=-30, lat_max=30)

        Plotter([data], xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                ymin=-90, ymax=90, coastlinescolour='red', layer_opts=[{'itemwidth': 50}])

        self.check_graphic()

    @skip("I can't make this plot 0-360 whatever I try...")
    def test_other_longitude_wrapping_multiple_ranges_forced_0_to_360(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.plotting.formatted_plot import Plotter
        from cis.test.util.mock import make_regular_2d_ungridded_data

        datas = [make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.),
                 make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=30, lat_max=60)]

        Plotter(datas, xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                ymin=-90, ymax=90, xmin=-180, xmax=180,
                layer_opts=[{'label': '-180to180', 'itemwidth': 50}, {'label': '0to360', 'itemwidth': 50}])

        self.check_graphic()

    def test_other_longitude_wrapping_multiple_ranges_forced_minus_180_to_180(self):
        """
        Test that ungridded data which crosses the dateline gets plotted correctly
        """
        from cis.plotting.formatted_plot import Plotter
        from cis.test.util.mock import make_regular_2d_ungridded_data

        datas = [make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=90, lon_min=-175., lon_max=145.),
                 make_regular_2d_ungridded_data(lon_dim_length=90, lon_min=5., lon_max=325., lat_min=30, lat_max=60)]

        Plotter(datas, xaxis='longitude', yaxis='latitude', width=8, height=6, cbarscale=None,
                ymin=-90, ymax=90, xmin=-180, xmax=180,
                layer_opts=[{'label': '-180to180', 'itemwidth': 50}, {'label': '0to360', 'itemwidth': 50}])

        self.check_graphic()

    def test_iris_contour_over_heatmap(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--width 20 --height 15 --cbarscale 0.5".split()

        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":type=heatmap,cbarorient=vertical",
                     "solarupclear:" + escape_colons(valid_2d_filename) + ":type=contour,color=white,contlevels=[1,10,25,50,175],cbarorient=vertical"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_contour_over_heatmap_binary_cmap(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--xmin -180 --xmax 180 --width 20 --height 15 --cbarscale 0.5".split()

        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":type=heatmap,cmap=binary,cbarorient=vertical",
                     "solarupclear:" + escape_colons(valid_2d_filename) + ":type=contour,cmap=jet,contlevels=[1,10,25,50,175],cbarorient=vertical"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    @skip("For some reason matplotlib doesn't respect vmin... But you can just set contlevels instead.")
    def test_transparent_contour_over_bluemarble(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type contourf --xmin -180 --xmax 180 --width 20 --height 15 --cbarscale 0.5" \
               " --nasabluemarble".split()

        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":cmap=Reds,alpha=0.5,vmin=0.000075"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    @skip("The wrapping here isn't working for some reason")
    def test_filled_contour_over_scatter2d(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--type contourf --width 20 --height 15 --xaxis longitude --yaxis latitude " \
               " --ymin 0 --ymax 90".split()

        arguments = ["plot", "GGALT:" + escape_colons(valid_NCAR_NetCDF_RAF_filename) + ":type=scatter2d,itemwidth=20",
                     "solarupclear:" + escape_colons(valid_2d_filename) + ":contlevels=[0,10,20,30,40,50,100],alpha=0.7,"
                                              "contlabel=true"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    @skip("The wrapping here isn't working for some reason")
    def test_filled_contour_over_scatter2d_with_cmin(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--width 20 --height 15 --xaxis longitude --yaxis latitude --xmin -180" \
               " --xmax -90 --ymin 0 --ymax 90 --nasabluemarble".split()

        arguments = ["plot", "GGALT:" + escape_colons(valid_NCAR_NetCDF_RAF_filename) + ":type=scatter2d,itemwidth=20",
                     "solarupclear:" + escape_colons(valid_2d_filename) + ":type=contourf,contlevels=[40,50,100],alpha=0.3,contlabel=true,"
                                              "cmap=Reds"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_scatter2d_over_contour(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--xlabel overiddenxlabel --height 10 --width 12 --xmin 0 --xmax 200 --xstep 10" \
               " --cbarorient horizontal --ymin 0 --ymax 90 --vmin 0 --cbarorient horizontal".split() + \
               ["--title=Overlay test"]
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":type=contour" ,
                     "snow:" + escape_colons(valid_2d_filename) + ":type=scatter2d,itemstyle=^,label=snowlabel,itemwidth=3"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_iris_scatter2d_overlay(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = "--xlabel overiddenxlabel --height 10 --width 12 --xmin 0 --xmax 200 --xstep 10" \
               " --cbarorient horizontal --ymin 0 --ymax 90 --vmin 0 --cbarorient horizontal".split() + \
               ["--title=Overlay test"]
        arguments = ["plot", "rain:" + escape_colons(valid_2d_filename) + ":type=heatmap" ,
                     "snow:" + escape_colons(valid_2d_filename) + ":type=scatter2d,itemstyle=^,label=snowlabel,itemwidth=3"]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_aerosol_cci_default_axes(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = []

        arguments = ["plot", valid_aerosol_cci_variable + ":" + escape_colons(valid_aerosol_cci_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_polar_projection(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = ["--projection=NorthPolarStereo"]

        arguments = ["plot", valid_aerosol_cci_variable + ":" + escape_colons(valid_aerosol_cci_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_mercator_projection(self):
        output_file_opt = ["--output", self.id() + ".png"]
        # These limits are ignored
        opts = ["--projection=Mercator", "--xmin=-90", "--xmax=90", "--nasabluemarble"]

        arguments = ["plot", valid_aerosol_cci_variable + ":" + escape_colons(valid_aerosol_cci_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_setting_xrange_using_datetimes(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = ["--xmin=2003-08-01",  "--xmax=2003-12-01T12:30:12"]

        arguments = ["plot", valid_aeronet_variable + ":" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_aeronet_default_axes(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = []

        arguments = ["plot", valid_aeronet_variable + ":" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_multiple_time_series_default_axes(self):
        # JASCIS-231
        output_file_opt = ["--output", self.id() + ".png"]
        opts = []

        # This guesses the axes based on the first datagroup, and then the second one based on the shape.
        arguments = ["plot", valid_aeronet_variable + ":" + escape_colons(another_valid_aeronet_filename),
                     'AOT_440:' + escape_colons(aggregated_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_multiple_time_series_incompatible_axes(self):
        # JASCIS-231
        output_file_opt = ["--output", self.id() + ".png"]
        opts = ['--type=line']

        # The aggregated data has some guessed axis labels (x and y are lon and lat respectively) which don't
        #  correspond to the other aeronet file, the plotter should ignore these though since they are scalar coords.
        arguments = ["plot", 'AOT_440_std_dev:' + escape_colons(aggregated_aeronet_filename),
                     valid_aeronet_variable + ":" + escape_colons(another_valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_multiple_time_series_default_axes_files_with_named_xaxis(self):
        # JASCIS-231
        output_file_opt = ["--output", self.id() + ".png"]
        opts = ['--xaxis=time']

        arguments = ["plot", 'AOT_440_std_dev:' + escape_colons(aggregated_aeronet_filename),
                     valid_aeronet_variable + ":" + escape_colons(another_valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    def test_plotting_heatmap_of_aggregated_ungridded_data(self):
        output_file_opt = ["--output", self.id() + ".png"]
        # opts = ['--xmin=-0.5', '--xmax=360']
        opts = []

        arguments = ["plot", 'AOD550:' + escape_colons(make_pathname('aggregated_aerosol_cci.nc'))]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()

    @raises(ProductPluginException)
    def test_aeronet_multiple_variable_plots(self):
        output_file_opt = ["--output", self.id() + ".png"]
        opts = []

        # Plotting doesn't support multiple variables
        arguments = ["plot", "AOT_532,AOT_551" + ":" + escape_colons(valid_aeronet_filename)]

        main_arguments = parse_args(arguments + opts + output_file_opt)
        plot_cmd(main_arguments)

        self.check_graphic()


class TestPlotAPIVisual(VisualTest):

    def test_implicit_comparative_scatter(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.data_io.ungridded_data import UngriddedDataList

        d = UngriddedDataList([make_regular_2d_ungridded_data(), make_regular_2d_ungridded_data(data_offset=2)])
        # This is needed to setup the coord shapes unfortunately...
        _ = d[0].data
        _ = d[1].data
        d[0].metadata._name = 'snow'
        d[1].plot(xaxis=d[0])

        self.check_graphic()

    def test_explicit_comparative_scatter(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from cis.data_io.ungridded_data import UngriddedDataList

        d = UngriddedDataList([make_regular_2d_ungridded_data(), make_regular_2d_ungridded_data(data_offset=2)])
        # This is needed to setup the coord shapes unfortunately...
        # TODO: Fix this in the Coord somewhere
        _ = d[0].data
        _ = d[1].data
        d[0].metadata._name = 'snow'
        d.plot(how='comparativescatter')

        self.check_graphic()

    def test_iris_multiple_scatter(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedDataList

        # This only works with one dimensional gridded data
        d = GriddedDataList([make_mock_cube(lat_dim_length=0), make_mock_cube(lat_dim_length=0, data_offset=2)])
        d[0].var_name = 'snow'
        d[1].var_name = 'rain'

        # Will default to line plots
        d.plot()

        self.check_graphic()

    def test_iris_comparative_scatter(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedDataList

        d = GriddedDataList([make_mock_cube(), make_mock_cube(data_offset=2)])
        d[0].var_name = 'snow'
        d[1].var_name = 'rain'

        d.plot(how='comparativescatter')

        self.check_graphic()

    def test_heatmap(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedData

        d = GriddedData.make_from_cube(make_mock_cube())
        d.plot()

        self.check_graphic()

    def test_contour_over_bluemarble(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedData

        d = GriddedData.make_from_cube(make_mock_cube())
        ax = d.plot(how='contour')
        ax.bluemarble()

        self.check_graphic()

    def test_multiple_line(self):
        from cis.test.util.mock import make_dummy_ungridded_data_time_series
        from cis.data_io.ungridded_data import UngriddedDataList

        d = UngriddedDataList([make_dummy_ungridded_data_time_series(), make_dummy_ungridded_data_time_series()])
        _ = d[0].data
        d[1].data += 2.0
        d[1].metadata._name = 'snow'

        d.plot(how='line')

        self.check_graphic()

    def test_scatter2d_over_bluemarble(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        d = make_regular_2d_ungridded_data()
        ax = d.plot(how='scatter2d')

        ax.bluemarble()

        self.check_graphic()

    def test_scatter2d_over_bluemarble_explicit_axis(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        d = make_regular_2d_ungridded_data()
        d.lat.axis = ''
        d.lon.axis=''
        ax = d.plot(how='scatter2d', xaxis='longitude', yaxis='latitude')

        ax.bluemarble()

        self.check_graphic()

    def test_scatter2d_over_bluemarble_coord_axis(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        d = make_regular_2d_ungridded_data()
        d.lat.axis = ''
        d.lon.axis=''
        ax = d.plot(how='scatter2d', xaxis=d.lon, yaxis=d.lat)

        ax.bluemarble()

        self.check_graphic()

    def test_mpl_kwargs(self):
        from cis.test.util.mock import make_dummy_ungridded_data_time_series

        d = make_dummy_ungridded_data_time_series()

        d.plot(how='line', c='purple', itemstyle='dashed')

        self.check_graphic()

    def test_layer_opts(self):
        from cis.test.util.mock import make_dummy_ungridded_data_time_series
        from cis.data_io.ungridded_data import UngriddedDataList

        d = UngriddedDataList([make_dummy_ungridded_data_time_series(),
                               make_dummy_ungridded_data_time_series()])
        _ = d[0].data
        d[1].data += 2.0
        d[1].metadata._name = 'snow'

        d.plot(how='line', layer_opts=[dict(c='yellow', itemstyle='dotted'),
                                       dict(c='purple', itemstyle='dashed')])

        self.check_graphic()

    def test_histogram(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        d = make_regular_2d_ungridded_data()
        d.plot(how='histogram')

        self.check_graphic()

    def test_histogram_2d(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedDataList

        d = GriddedDataList([make_mock_cube(), make_mock_cube(data_offset=2)])
        d[0].var_name = 'snow'
        d[1].var_name = 'rain'

        d.plot(how='histogram2d')

        self.check_graphic()

    def test_invalid_args(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data
        from nose.tools import assert_raises

        d = make_regular_2d_ungridded_data()

        with assert_raises(AttributeError):
            d.plot(how='histogram', contnlevels=5)

        #TODO: There must be more of these

    def test_taylor_diagram_gridded(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import GriddedDataList

        d = GriddedDataList([make_mock_cube(), make_mock_cube(data_offset=2)])
        d[0].var_name = 'snow'
        d[1].var_name = 'rain'

        d.plot(how='taylor')

        self.check_graphic()

    def test_orographic_projection(self):
        from cis import read_data
        import cartopy.crs as ccrs

        d = read_data(valid_aerosol_cci_filename, valid_aerosol_cci_variable)

        ax = plt.subplot(1, 1, 1, projection=ccrs.Orthographic(0, 90))

        ax = d.plot(ax=ax)

        ax.set_global()

        self.check_graphic()

    def test_polar_projection(self):
        from cis import read_data
        import cartopy.crs as ccrs

        d = read_data(valid_aerosol_cci_filename, valid_aerosol_cci_variable)

        ax = d.plot(projection=ccrs.NorthPolarStereo())

        self.check_graphic()

    def test_mercator_projection(self):
        from cis import read_data
        import cartopy.crs as ccrs

        d = read_data(valid_aerosol_cci_filename, valid_aerosol_cci_variable)

        ax = d.plot(projection=ccrs.Mercator())

        ax.bluemarble()

        self.check_graphic()

    def test_can_specify_yaxis_altitude(self):
        from cis import read_data
        d = read_data(valid_GASSP_aeroplane_filename, valid_GASSP_aeroplane_variable)
        d.plot(yaxis='altitude')

        self.check_graphic()

    def test_log_colorbar_scale(self):
        from cis import read_data
        d = read_data(valid_aerosol_cci_filename, valid_aerosol_cci_variable)
        d.plot(logv=True)

        self.check_graphic()




"""
Custom CIS exceptions
"""


class CISError(Exception):
    pass


class InvalidPlotTypeError(CISError):
    pass


class InvalidDimensionError(CISError):
    pass


class InvalidVariableError(CISError):
    pass


class InconsistentDimensionsError(CISError):
    pass


class InvalidPlotFormatError(CISError):
    pass


class InvalidFileExtensionError(CISError):
    pass


class InvalidDataTypeError(CISError):
    pass


class InvalidOperationError(CISError):
    pass


class InvalidLineStyleError(CISError):
    pass


class InvalidHistogramStyleError(CISError):
    pass


class CoordinateNotFoundError(CISError):
    pass


class DuplicateCoordinateError(CISError):
    pass


class ClassNotFoundError(CISError):
    pass


class InvalidCommandLineOptionError(CISError):
    pass


class InvalidNumberOfDatagroupsSpecifiedError(CISError):
    pass


class NotEnoughAxesSpecifiedError(CISError):
    pass


class NoDataInSubsetError(CISError):
    pass


class FileFormatError(CISError):
    """
    Throw when there is an error determining the type of a file
    """
    error_list = ['Unknown error']

    def __init__(self, error_list, *args, **kwargs):
        super(FileFormatError, self).__init__(args, kwargs)
        self.error_list = error_list


class UserPrintableException(CISError):
    """
    This exception is thrown if the program has failed for a known reason. This message is printed without a stack trace
    """

    def __init__(self, message):
        """
        Constructor
        :param message: the message to show the user
        :return: nothing
        """
        super(UserPrintableException, self).__init__()
        self.message = message

    def __str__(self):
        return self.message



import logging

import iris
import iris.coords
from iris.exceptions import CoordinateMultiDimError
import numpy as np
from numpy import mean as np_mean, std as np_std, min as np_min, max as np_max, sum as np_sum

from cis.collocation.col_framework import (Collocator, Constraint, PointConstraint, CellConstraint,
                                           IndexedConstraint, Kernel, AbstractDataOnlyKernel)
import cis.exceptions
from cis.data_io.gridded_data import GriddedData, make_from_cube, GriddedDataList
from cis.data_io.hyperpoint import HyperPoint, HyperPointList
from cis.data_io.ungridded_data import Metadata, UngriddedDataList, UngriddedData
import cis.collocation.data_index as data_index
from cis.utils import log_memory_profile, set_standard_name_if_valid


class GeneralUngriddedCollocator(Collocator):
    """
    Collocator for locating onto ungridded sample points
    """

    def collocate(self, points, data, constraint, kernel):
        """
        This collocator takes a list of HyperPoints and a data object (currently either Ungridded
        data or a Cube) and returns one new LazyData object with the values as determined by the
        constraint and kernel objects. The metadata for the output LazyData object is copied from
        the input data object.

        :param UngriddedData or UngriddedCoordinates points: Object defining the sample points
        :param UngriddedData data: The source data to collocate from
        :param constraint: An instance of a Constraint subclass which takes a data object and
                           returns a subset of that data based on it's internal parameters
        :param kernel: An instance of a Kernel subclass which takes a number of points and returns
                       a single value
        :return UngriddedData or UngriddedDataList: Depending on the input
        """
        log_memory_profile("GeneralUngriddedCollocator Initial")

        if isinstance(data, list):
            # Indexing and constraints (for SepConstraintKdTree) will only take place on the first iteration,
            # so we really can just call this method recursively if we've got a list of data.
            output = UngriddedDataList()
            for var in data:
                output.extend(self.collocate(points, var, constraint, kernel))
            return output

        # First fix the sample points so that they all fall within the same 360 degree longitude range
        _fix_longitude_range(points.coords(), points)
        # Then fix the data points so that they fall onto the same 360 degree longitude range as the sample points
        _fix_longitude_range(points.coords(), data)

        # Convert to dataframes for fancy indexing
        sample_points = points.as_data_frame(time_index=False, name='vals')
        data_points = data.as_data_frame(time_index=False, name='vals').dropna(axis=0)

        log_memory_profile("GeneralUngriddedCollocator after data retrieval")

        # Create index if constraint and/or kernel require one.
        coord_map = None
        data_index.create_indexes(constraint, points, data_points, coord_map)
        log_memory_profile("GeneralUngriddedCollocator after indexing")

        logging.info("--> Collocating...")

        # Create output arrays.
        self.var_name = data.var_name
        self.var_long_name = data.long_name
        self.var_standard_name = data.standard_name
        self.var_units = data.units
        var_set_details = kernel.get_variable_details(self.var_name, self.var_long_name,
                                                      self.var_standard_name, self.var_units)

        sample_points_count = len(sample_points)
        # Create an empty masked array to store the collocated values. The elements will be unmasked by assignment.
        values = np.ma.masked_all((len(var_set_details), sample_points_count))
        values.fill_value = self.fill_value
        log_memory_profile("GeneralUngriddedCollocator after output array creation")

        logging.info("    {} sample points".format(sample_points_count))
        # Apply constraint and/or kernel to each sample point.

        if isinstance(kernel, nn_horizontal_only):
            # Only find the nearest point using the kd-tree, without constraint in other dimensions
            nearest_points = data_points.iloc[constraint.haversine_distance_kd_tree_index.find_nearest_point(sample_points)]
            values[0, :] = nearest_points.vals.values
        else:
            for i, point, con_points in constraint.get_iterator(self.missing_data_for_missing_sample, None, None,
                                                                data_points, None, sample_points, None):

                try:
                    values[:, i] = kernel.get_value(point, con_points)
                    # Kernel returns either a single value or a tuple of values to insert into each output variable.
                except CoordinateMultiDimError as e:
                    raise NotImplementedError(e)
                except ValueError as e:
                    pass
        log_memory_profile("GeneralUngriddedCollocator after running kernel on sample points")

        # Mask any bad values
        values = np.ma.masked_invalid(values)

        return_data = UngriddedDataList()
        for idx, var_details in enumerate(var_set_details):
            var_metadata = Metadata(name=var_details[0], long_name=var_details[1], shape=(len(sample_points),),
                                    missing_value=self.fill_value, units=var_details[3])
            set_standard_name_if_valid(var_metadata, var_details[2])
            return_data.append(UngriddedData(values[idx, :], var_metadata, points.coords()))
        log_memory_profile("GeneralUngriddedCollocator final")

        return return_data


class GriddedUngriddedCollocator(Collocator):
    """
    Collocator for locating GriddedData onto ungridded sample points
    """

    def __init__(self, fill_value=None, var_name='', var_long_name='', var_units='',
                 missing_data_for_missing_sample=False, extrapolate=False):
        super(GriddedUngriddedCollocator, self).__init__(fill_value, var_name, var_long_name, var_units,
                                                         missing_data_for_missing_sample)
        self.extrapolate = extrapolate
        self.interpolator = None

    def collocate(self, points, data, constraint, kernel):
        """
        This collocator takes a list of HyperPoints and a data object (currently either Ungridded
        data or a Cube) and returns one new LazyData object with the values as determined by the
        constraint and kernel objects. The metadata for the output LazyData object is copied from
        the input data object.

        :param UngriddedData or UngriddedCoordinates points: Objects defining the sample points
        :param GriddedData or GriddedDataList data: Data to resample
        :param constraint: An instance of a Constraint subclass which takes a data object and
                           returns a subset of that data based on it's internal parameters
        :param kernel: An instance of a Kernel subclass which takes a number of points and returns
                       a single value
        :return: A single LazyData object
        """
        from cis.collocation.gridded_interpolation import GriddedUngriddedInterpolator
        log_memory_profile("GriddedUngriddedCollocator Initial")

        if isinstance(data, list):
            # Indexing and constraints (for SepConstraintKdTree) will only take place on the first iteration,
            # so we really can just call this method recursively if we've got a list of data.
            output = UngriddedDataList()
            for var in data:
                output.extend(self.collocate(points, var, constraint, kernel))
            return output

        if constraint is not None and not isinstance(constraint, DummyConstraint):
            raise ValueError("A constraint cannot be specified for the GriddedUngriddedCollocator")
        data_points = data

        # First fix the sample points so that they all fall within the same 360 degree longitude range
        _fix_longitude_range(points.coords(), points)
        # Then fix the data points so that they fall onto the same 360 degree longitude range as the sample points
        _fix_longitude_range(points.coords(), data_points)

        log_memory_profile("GriddedUngriddedCollocator after data retrieval")

        logging.info("--> Collocating...")
        logging.info("    {} sample points".format(points.size))

        if self.interpolator is None:
            # Cache the interpolator
            self.interpolator = GriddedUngriddedInterpolator(data, points, kernel, self.missing_data_for_missing_sample)

        values = self.interpolator(data, fill_value=self.fill_value, extrapolate=self.extrapolate)

        log_memory_profile("GriddedUngriddedCollocator after running kernel on sample points")

        metadata = Metadata(self.var_name or data.var_name, long_name=self.var_long_name or data.long_name,
                            shape=values.shape, missing_value=self.fill_value, units=self.var_units or data.units)
        set_standard_name_if_valid(metadata, data.standard_name)
        return_data = UngriddedDataList([UngriddedData(values, metadata, points.coords())])

        log_memory_profile("GriddedUngriddedCollocator final")

        return return_data


class DummyCollocator(Collocator):
    def collocate(self, points, data, constraint, kernel):
        """
            This collocator does no collocation at all - it just returns the original data values. This might be useful
            if the input data for one variable is already known to be on the same grid as points. This routine could
            check the coordinates are the same but currently does no such check.

        :param points: A list of HyperPoints
        :param data: An UngriddedData object or Cube
        :param constraint: Unused
        :param kernel: Unused
        :return: A single LazyData object
        """
        from cis.data_io.ungridded_data import LazyData

        logging.info("--> Collocating...")
        data = cis.utils.listify(data)
        return [LazyData(var.data, var.metadata) for var in data]


class DummyConstraint(Constraint):
    def constrain_points(self, point, data):
        # This is a null constraint - all of the points just get passed back
        return data


class SepConstraint(PointConstraint):
    def __init__(self, h_sep=None, a_sep=None, p_sep=None, t_sep=None):
        from cis.exceptions import InvalidCommandLineOptionError

        super(SepConstraint, self).__init__()

        self.checks = []

        if h_sep is not None:
            self.h_sep = cis.utils.parse_distance_with_units_to_float_km(h_sep)
            self.checks.append(self.horizontal_constraint)
        if a_sep is not None:
            self.a_sep = cis.utils.parse_distance_with_units_to_float_m(a_sep)
            self.checks.append(self.alt_constraint)
        if p_sep is not None:
            try:
                self.p_sep = float(p_sep)
            except:
                raise InvalidCommandLineOptionError('Separation Constraint p_sep must be a valid float')
            self.checks.append(self.pressure_constraint)
        if t_sep is not None:
            from cis.parse_datetime import parse_datetimestr_delta_to_float_days
            try:
                self.t_sep = parse_datetimestr_delta_to_float_days(t_sep)
            except ValueError as e:
                raise InvalidCommandLineOptionError(e)
            self.checks.append(self.time_constraint)

    def time_constraint(self, point, ref_point):
        return point.time_sep(ref_point) < self.t_sep

    def alt_constraint(self, point, ref_point):
        return point.alt_sep(ref_point) < self.a_sep

    def pressure_constraint(self, point, ref_point):
        return point.pres_sep(ref_point) < self.p_sep

    def horizontal_constraint(self, point, ref_point):
        return point.haversine_dist(ref_point) < self.h_sep

    def constrain_points(self, ref_point, data):
        con_points = HyperPointList()
        for point in data:
            if all(check(point, ref_point) for check in self.checks):
                con_points.append(point)
        return con_points


class SepConstraintKdtree(PointConstraint):
    """A separation constraint that uses a k-D tree to optimise spatial constraining.
    If no horizontal separation parameter is supplied, this reduces to an exhaustive
    search using the other parameter(s).
    """

    def __init__(self, h_sep=None, a_sep=None, p_sep=None, t_sep=None):
        from cis.exceptions import InvalidCommandLineOptionError

        self.haversine_distance_kd_tree_index = False

        super(SepConstraintKdtree, self).__init__()

        self._index_cache = {}
        self.checks = []
        if h_sep is not None:
            self.h_sep = cis.utils.parse_distance_with_units_to_float_km(h_sep)
            self.haversine_distance_kd_tree_index = None
        else:
            self.h_sep = None

        if a_sep is not None:
            self.a_sep = cis.utils.parse_distance_with_units_to_float_m(a_sep)
            self.checks.append(self.alt_constraint)
        if p_sep is not None:
            try:
                self.p_sep = float(p_sep)
            except:
                raise InvalidCommandLineOptionError('Separation Constraint p_sep must be a valid float')
            self.checks.append(self.pressure_constraint)
        if t_sep is not None:
            from cis.parse_datetime import parse_datetimestr_delta_to_float_days
            try:
                self.t_sep = parse_datetimestr_delta_to_float_days(t_sep)
            except ValueError as e:
                raise InvalidCommandLineOptionError(e)
            self.checks.append(self.time_constraint)

    def time_constraint(self, points, ref_point):
        return (np.abs(points.time - ref_point.time) < self.t_sep).to_numpy().nonzero()[0]

    def alt_constraint(self, points, ref_point):
        return (np.abs(points.altitude - ref_point.altitude) < self.a_sep).to_numpy().nonzero()[0]

    def pressure_constraint(self, points, ref_point):
        greater_pressures = np.nonzero(((points.air_pressure.values / ref_point.air_pressure) < self.p_sep) &
                                       (points.air_pressure.values > ref_point.air_pressure))[0]
        lesser_pressures = np.nonzero(((ref_point.air_pressure / points.air_pressure.values) < self.p_sep) &
                                      (points.air_pressure.values <= ref_point.air_pressure))[0]
        return np.concatenate([lesser_pressures, greater_pressures])

    def constrain_points(self, ref_point, data):
        if self.haversine_distance_kd_tree_index and self.h_sep:
            point_indices = self._get_cached_indices(ref_point)
            if point_indices is None:
                point_indices = self.haversine_distance_kd_tree_index.find_points_within_distance(ref_point, self.h_sep)
                self._add_cached_indices(ref_point, point_indices)
            con_points = data.iloc[point_indices]
        else:
            con_points = data
        for check in self.checks:
            con_points = con_points.iloc[check(con_points, ref_point)]

        return con_points

    def _get_cached_indices(self, ref_point):
        # Don't use the value as a key (it's both irrelevant and un-hashable)
        return self._index_cache.get(tuple(ref_point[['latitude', 'longitude']].values), None)

    def _add_cached_indices(self, ref_point, indices):
        # Don't use the value as a key (it's both irrelevant and un-hashable)
        self._index_cache[tuple(ref_point[['latitude', 'longitude']].values)] = indices

    def get_iterator(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, output_data):
        cell_count = 0
        total_count = 0
        sample_points_count = len(points)

        indices = False

        if self.haversine_distance_kd_tree_index and self.h_sep:
            indices = self.haversine_distance_kd_tree_index.find_points_within_distance_sample(points, self.h_sep)

        for i, p in points.iterrows():

            # Log progress periodically.
            cell_count += 1
            if cell_count == 1000:
                total_count += cell_count
                cell_count = 0
                logging.info("    Processed {} points of {}".format(total_count, sample_points_count))

            # If missing_data_for_missing_sample
            if not (missing_data_for_missing_sample and (hasattr(p, 'vals') and np.isnan(p.vals))):
                if indices:
                    # Note that data_points has to be a dataframe at this point because of the indexing
                    d_points = data_points.iloc[indices[i]]
                else:
                    d_points = data_points
                for check in self.checks:
                    d_points = d_points.iloc[check(d_points, p)]

                yield i, p, d_points


# noinspection PyPep8Naming
class mean(AbstractDataOnlyKernel):
    """
    Calculate mean of data points
    """

    def get_value_for_data_only(self, values):
        """
        return the mean
        """
        return np_mean(values)


# noinspection PyPep8Naming
class stddev(AbstractDataOnlyKernel):
    """
    Calculate the standard deviation
    """

    def get_value_for_data_only(self, values):
        """
        Return the standard deviation points
        """
        return np_std(values, ddof=1)


# noinspection PyPep8Naming,PyShadowingBuiltins
class min(AbstractDataOnlyKernel):
    """
    Calculate the minimum value
    """

    def get_value_for_data_only(self, values):
        """
        Return the minimum value
        """
        return np_min(values)


# noinspection PyPep8Naming,PyShadowingBuiltins
class max(AbstractDataOnlyKernel):
    """
    Calculate the maximum value
    """

    def get_value_for_data_only(self, values):
        """
        Return the maximum value
        """
        return np_max(values)


class sum(AbstractDataOnlyKernel):
    """
    Calculate the sum of the values
    """

    def get_value_for_data_only(self, values):
        """
        Return the sum of the values
        """
        return np_sum(values)


# noinspection PyPep8Naming
class moments(AbstractDataOnlyKernel):
    return_size = 3

    def __init__(self, mean_name='', stddev_name='', nopoints_name=''):
        self.mean_name = mean_name
        self.stddev_name = stddev_name
        self.nopoints_name = nopoints_name

    def get_variable_details(self, var_name, var_long_name, var_standard_name, var_units):
        """Sets name and units for mean, standard deviation and number of points variables, based
        on those of the base variable or overridden by those specified as kernel parameters.
        :param var_name: base variable name
        :param var_long_name: base variable long name
        :param var_standard_name: base variable standard name
        :param var_units: base variable units
        :return: tuple of tuples each containing (variable name, variable long name, variable units)
        """
        self.mean_name = var_name
        self.stddev_name = var_name + '_std_dev'
        stdev_long_name = 'Corrected sample standard deviation of %s' % var_long_name
        stddev_units = var_units
        self.nopoints_name = var_name + '_num_points'
        npoints_long_name = 'Number of points used to calculate the mean of %s' % var_long_name
        npoints_units = ''
        return ((self.mean_name, var_long_name, var_standard_name, var_units),
                (self.stddev_name, stdev_long_name, None, stddev_units),
                (self.nopoints_name, npoints_long_name, None, npoints_units))

    def get_value_for_data_only(self, values):
        """
        Returns the mean, standard deviation and number of values
        """

        return np_mean(values), np_std(values, ddof=1), np.size(values)


class nn_horizontal(Kernel):
    def get_value(self, point, data):
        """
            Collocation using nearest neighbours along the face of the earth where both points and
              data are a list of HyperPoints. The default point is the first point.
        """
        from cis.collocation.kdtree import haversine
        iterator = data.iterrows()
        try:
            nearest_point = next(iterator)[1]
        except StopIteration:
            # No points to check
            raise ValueError
        for idx, data_point in iterator:
            if (haversine(point, nearest_point) > haversine(point, data_point)):
                nearest_point = data_point
        return nearest_point.vals


class nn_horizontal_only(Kernel):
    """
    Used to shortcut the constraint checks in dimensions other than the horizontal to allow quick kd-tree lookup
    """
    def get_value(self, point, data):
        pass


class nn_altitude(Kernel):
    def get_value(self, point, data):
        """
            Collocation using nearest neighbours in altitude, where both points and
              data are a list of HyperPoints. The default point is the first point.
        """
        iterator = data.iterrows()
        try:
            nearest_point = next(iterator)[1]
        except StopIteration:
            # No points to check
            raise ValueError
        for idx, data_point in iterator:
            if abs(point.altitude - nearest_point.altitude) > abs(point.altitude - data_point.altitude):
                nearest_point = data_point
        return nearest_point.vals


class nn_pressure(Kernel):


    def get_value(self, point, data):
        """
            Collocation using nearest neighbours in pressure, where both points and
              data are a list of HyperPoints. The default point is the first point.
        """

        def pres_sep(point1, point2):
            """
                Computes the pressure ratio between two points, this is always >= 1.
            """
            if point1.air_pressure > point2.air_pressure:
                return point1.air_pressure / point2.air_pressure
            else:
                return point2.air_pressure / point1.air_pressure

        iterator = data.iterrows()
        try:
            nearest_point = next(iterator)[1]
        except StopIteration:
            # No points to check
            raise ValueError
        for idx, data_point in iterator:
            if pres_sep(point, nearest_point) > pres_sep(point, data_point):
                nearest_point = data_point
        return nearest_point.vals


class nn_time(Kernel):
    def get_value(self, point, data):
        """
            Collocation using nearest neighbours in time, where both points and
              data are a list of HyperPoints. The default point is the first point.
        """
        iterator = data.iterrows()
        try:
            nearest_point = next(iterator)[1]
        except StopIteration:
            # No points to check
            raise ValueError
        for idx, data_point in iterator:
            if abs(point.time - nearest_point.time) > abs(point.time - data_point.time):
                nearest_point = data_point
        return nearest_point.vals


# These classes act as abbreviations for kernel classes above:
class nn_h(nn_horizontal):
    """Nearest neighbour horizontal kernel - alias for nn_horizontal.
    """
    pass


class nn_a(nn_altitude):
    """Nearest neighbour altitude kernel - alias for nn_altitude.
    """
    pass


class nn_p(nn_pressure):
    """Nearest neighbour pressure kernel - alias for nn_pressure.
    """
    pass


class nn_t(nn_time):
    """Nearest neighbour time kernel - alias for nn_time.
    """
    pass


class GriddedCollocator(Collocator):

    def __init__(self, fill_value=None, var_name='', var_long_name='', var_units='',
                 missing_data_for_missing_sample=False, extrapolate=False):
        super(GriddedCollocator, self).__init__(fill_value, var_name, var_long_name, var_units,
                                                         missing_data_for_missing_sample)
        self.extrapolate = 'extrapolate' if extrapolate else 'mask'

    @staticmethod
    def _check_for_valid_kernel(kernel):
        from cis.exceptions import ClassNotFoundError

        if not (isinstance(kernel, gridded_gridded_nn) or isinstance(kernel, gridded_gridded_li)):
            raise ClassNotFoundError("Expected kernel of one of classes {}; found one of class {}".format(
                str([cis.utils.get_class_name(gridded_gridded_nn),
                     cis.utils.get_class_name(gridded_gridded_li)]),
                cis.utils.get_class_name(type(kernel))))

    def collocate(self, points, data, constraint, kernel):
        """
        This collocator takes two Iris cubes, and collocates from the data cube onto the grid of the 'points' cube. The
        collocator then returns another Iris cube.
        :param points: An Iris cube with the sampling grid to collocate onto.
        :param data: The Iris cube with the data to be collocated.
        :param constraint: None allowed yet, as this is unlikely to be required for gridded-gridded.
        :param kernel: The kernel to use, current options are gridded_gridded_nn and gridded_gridded_li.
        :return: An Iris cube with the collocated data.
        """
        self._check_for_valid_kernel(kernel)

        # Force the data longitude range to be the same as that of the sample grid.
        _fix_longitude_range(points.coords(), data)

        # Initialise variables used to create an output mask based on the sample data mask.
        sample_coord_lookup = {}  # Maps coordinate in sample data -> location in dimension order
        for idx, coord in enumerate(points.coords()):
            sample_coord_lookup[coord] = idx
        sample_coord_transpose_map = []  # For coords in both sample and data, contains the position in the sample
        other_coord_transpose_map = []  # For coords in data but not in sample, records that coord's position in data.
        repeat_size = 1
        output_mask = np.ma.nomask

        # Make a list of the coordinates we have, with each entry containing a list with the name of the coordinate and
        # the number of points along its axis. One is for the sample grid, which contains the points where we
        # interpolate too, and one is for the output grid, which will additionally contain any dimensions missing in the
        # sample grid.
        coord_names_and_sizes_for_sample_grid = []
        coord_names_and_sizes_for_output_grid = []
        for idx, coord in enumerate(data.coords(dim_coords=True)):
            # First try and find the coordinate in points, the sample grid. If an exception is thrown, it means that
            # name does not appear in the sample grid, and instead take the coordinate name and length from the original
            # data, as this is what we will be keeping.
            try:
                sample_coord = points.coords(coord.name())[0]
                coord_names_and_sizes_for_sample_grid.append([coord.name(), len(sample_coord.points)])
                # Find the index of the sample coordinate corresponding to the data coordinate.
                sample_coord_transpose_map.append(sample_coord_lookup[sample_coord])
            except IndexError:
                coord_names_and_sizes_for_output_grid.append([coord.name(), len(coord.points)])
                repeat_size *= len(coord.points)
                other_coord_transpose_map.append(idx)

        # Now we sort the sample coordinates so that they are in the same order as in the sample file,
        # rather than the order of the data file (that's the order we want the output dimensions).
        coord_names_and_sizes_for_sample_grid = [x[0] for x in sorted(zip(coord_names_and_sizes_for_sample_grid,
                                                                          sample_coord_transpose_map),
                                                                      key=lambda t: t[1])]

        # Adding the lists together in this way ensures that the coordinates not in the sample grid appear in the final
        # position, which is important for adding the points from the Iris interpolater to the new array. The data
        # returned from the Iris interpolater method will have dimensions of these missing coordinates, which needs
        # to be the final dimensions in the numpy array, as the iterator will give the position of the other dimensions.
        coord_names_and_sizes_for_output_grid = coord_names_and_sizes_for_sample_grid + \
                                                coord_names_and_sizes_for_output_grid

        # An array for the collocated data, with the correct shape
        output_shape = tuple(i[1] for i in coord_names_and_sizes_for_output_grid)
        new_data = np.zeros(output_shape)

        if self.missing_data_for_missing_sample:
            output_mask = self._make_output_mask(coord_names_and_sizes_for_sample_grid, output_shape,
                                                 points, repeat_size)

        # Now recreate the points cube, while ignoring any DimCoords in points that are not in the data cube
        new_dim_coord_list = []
        new_points_array_shape = []
        for i in range(0, len(coord_names_and_sizes_for_output_grid)):
            # Try and find the coordinate in the sample grid
            coord_found = points.coords(coord_names_and_sizes_for_output_grid[i][0])

            # If the coordinate exists in the sample grid then append the new coordinate to the list. Iris requires
            # this be given as a DimCoord object, along with a axis number, in a tuple pair.
            if len(coord_found) != 0:
                new_dim_coord_list.append((coord_found[0], len(new_dim_coord_list)))
                new_points_array_shape.append(coord_found[0].points.size)

        new_points_array = np.zeros(tuple(new_points_array_shape))

        # Use the new_data array to recreate points, without the DimCoords not in the data cube
        points = iris.cube.Cube(new_points_array, dim_coords_and_dims=new_dim_coord_list)

        output_cube = self._iris_interpolate(coord_names_and_sizes_for_output_grid,
                                             coord_names_and_sizes_for_sample_grid, data,
                                             kernel, output_mask, points, self.extrapolate)

        if not isinstance(output_cube, list):
            return GriddedDataList([output_cube])
        else:
            return output_cube

    @staticmethod
    def _make_output_mask(coord_names_and_sizes_for_sample_grid, output_shape, points, repeat_size):
        """ Creates a mask to apply to the output data based on the sample data mask. If there are coordinates in
        the data grid that are not in the sample grid, the same mask value is repeated for all values of the
        extra coordinates. If there are coordinates in the sample grid that are not in the data grid, a mask
        is not created since there is many to one correspondence between sample and output grid points.
        """
        output_mask = None

        # Construct the missing value mask from the sample data, if applicable.
        if len(coord_names_and_sizes_for_sample_grid) < len(points.dim_coords):
            # One or more axes collapsed so many sample points correspond to each output point.
            pass
        else:
            input_mask = np.ma.getmask(points.data)
            if input_mask is np.ma.nomask:
                # No sample data missing-value mask.
                pass
            else:
                # Fill in the remaining coordinates (those from the data which are not in the sample) by repeating
                # the constructed mask for each value of those coordinates
                output_mask = np.reshape(np.repeat(points.data.mask, repeat_size), output_shape)
        return output_mask

    @staticmethod
    def _iris_interpolate(coord_names_and_sizes_for_output_grid, coord_names_and_sizes_for_sample_grid, data, kernel,
                          output_mask, points, extrapolate):
        """ Collocates using iris.analysis.interpolate
        """
        coordinate_point_pairs = []
        for j in range(0, len(coord_names_and_sizes_for_sample_grid)):
            # For each coordinate make the list of tuple pair Iris requires, for example
            # [('latitude', -90), ('longitude, 0')]
            coordinate_point_pairs.append((coord_names_and_sizes_for_sample_grid[j][0],
                                           points.dim_coords[j].points))

        # The result here will be a cube with the correct dimensions for the output, so interpolated over all points
        # in coord_names_and_sizes_for_output_grid.
        output_cube = make_from_cube(data.interpolate(coordinate_point_pairs,
                                                      kernel.interpolater(extrapolation_mode=extrapolate)))

        # Iris outputs interpolated cubes with the dimensions in the order of the data grid, not the sample grid,
        # so we need to rearrange the order of the dimensions.
        output_coord_lookup = {}
        for idx, coord in enumerate(output_cube.dim_coords):
            output_coord_lookup[coord.name()] = idx
        transpose_map = [output_coord_lookup[coord[0]] for coord in coord_names_and_sizes_for_output_grid]
        output_cube.transpose(transpose_map)

        if isinstance(output_cube, list):
            for idx, data in enumerate(output_cube):
                output_cube[idx].data = cis.utils.apply_mask_to_numpy_array(data.data, output_mask)
        else:
            output_cube.data = cis.utils.apply_mask_to_numpy_array(output_cube.data, output_mask)
        return output_cube


class gridded_gridded_nn(Kernel):
    def __init__(self):
        from iris.analysis import Nearest
        self.name = 'nearest'
        self.interpolater = Nearest

    def get_value(self, point, data):
        """Not needed for gridded/gridded collocation.
        """
        raise ValueError("gridded_gridded_nn kernel selected for use with collocator other than GriddedCollocator")


class gridded_gridded_li(Kernel):
    def __init__(self):
        from iris.analysis import Linear
        self.name = 'bilinear'
        self.interpolater = Linear

    def get_value(self, point, data):
        """Not needed for gridded/gridded collocation.
        """
        raise ValueError("gridded_gridded_li kernel selected for use with collocator other than GriddedCollocator")


class GeneralGriddedCollocator(Collocator):
    """Performs collocation of data on to the points of a cube (ie onto a gridded dataset).
    """

    def collocate(self, points, data, constraint, kernel):
        """
        :param points: cube defining the sample points
        :param data: CommonData object providing data to be collocated (or list of Data)
        :param constraint: instance of a Constraint subclass, which takes a data object and returns a subset of that
                           data based on it's internal parameters
        :param kernel: instance of a Kernel subclass which takes a number of points and returns a single value
        :return: GriddedDataList of collocated data
        """
        log_memory_profile("GeneralGriddedCollocator Initial")
        if isinstance(data, list):
            # If data is a list then call this method recursively over each element
            output_list = []
            for variable in data:
                collocated = self.collocate(points, variable, constraint, kernel)
                output_list.extend(collocated)
            return GriddedDataList(output_list)

        data_points = data.get_non_masked_points()

        log_memory_profile("GeneralGriddedCollocator Created data hyperpoint list view")

        # Work out how to iterate over the cube and map HyperPoint coordinates to cube coordinates.
        coord_map = make_coord_map(points, data)
        if self.missing_data_for_missing_sample and len(coord_map) is not len(points.coords()):
            raise cis.exceptions.UserPrintableException(
                "A sample variable has been specified but not all coordinates in the data appear in the sample so "
                "there are multiple points in the sample data so whether the data is missing or not can not be "
                "determined")

        coords = points.coords()
        shape = []
        output_coords = []

        # Find shape of coordinates to be iterated over.
        for (hpi, ci, shi) in coord_map:
            coord = coords[ci]
            if coord.ndim > 1:
                raise NotImplementedError("Co-location of data onto a cube with a coordinate of dimension greater"
                                          " than one is not supported (coordinate %s)", coord.name())
            # Ensure that bounds exist.
            if not coord.has_bounds():
                logging.warning("Creating guessed bounds as none exist in file")
                coord.guess_bounds()
            shape.append(coord.shape[0])
            output_coords.append(coord)

        _fix_longitude_range(coords, data_points)

        log_memory_profile("GeneralGriddedCollocator Created output coord map")

        # Create index if constraint supports it.
        data_index.create_indexes(constraint, coords, data_points, coord_map)
        data_index.create_indexes(kernel, points, data_points, coord_map)

        log_memory_profile("GeneralGriddedCollocator Created indexes")

        # Initialise output array as initially all masked, and set the appropriate fill value.
        values = []
        for i in range(kernel.return_size):
            val = np.ma.zeros(shape)
            val.mask = True
            val.fill_value = self.fill_value
            values.append(val)

        if kernel.return_size == 1:
            set_value_kernel = self._set_single_value_kernel
        else:
            set_value_kernel = self._set_multi_value_kernel

        logging.info("--> Co-locating...")

        if hasattr(kernel, "get_value_for_data_only") and hasattr(constraint, "get_iterator_for_data_only"):
            # Iterate over constrained cells
            iterator = constraint.get_iterator_for_data_only(
                self.missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, values)
            for out_indices, data_values in iterator:
                try:
                    kernel_val = kernel.get_value_for_data_only(data_values)
                    set_value_kernel(kernel_val, values, out_indices)
                except ValueError:
                    # ValueErrors are raised by Kernel when there are no points to operate on.
                    # We don't need to do anything.
                    pass
        else:
            # Iterate over constrained cells
            iterator = constraint.get_iterator(
                self.missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, values)
            for out_indices, hp, con_points in iterator:
                try:
                    kernel_val = kernel.get_value(hp, con_points)
                    set_value_kernel(kernel_val, values, out_indices)
                except ValueError:
                    # ValueErrors are raised by Kernel when there are no points to operate on.
                    # We don't need to do anything.
                    pass

        log_memory_profile("GeneralGriddedCollocator Completed collocation")

        # Construct an output cube containing the collocated data.
        kernel_var_details = kernel.get_variable_details(self.var_name or data.var_name,
                                                         self.var_long_name or data.long_name,
                                                         data.standard_name,
                                                         self.var_units or data.units)
        output = GriddedDataList([])
        for idx, val in enumerate(values):
            cube = self._create_collocated_cube(data, val, output_coords)
            data_with_nan_and_inf_removed = np.ma.masked_invalid(cube.data)
            data_with_nan_and_inf_removed.set_fill_value(self.fill_value)
            cube.data = data_with_nan_and_inf_removed
            cube.var_name = kernel_var_details[idx][0]
            cube.long_name = kernel_var_details[idx][1]
            set_standard_name_if_valid(cube, kernel_var_details[idx][2])
            try:
                cube.units = kernel_var_details[idx][3]
            except ValueError:
                logging.warn(
                    "Units are not cf compliant, not setting them. Units {}".format(kernel_var_details[idx][3]))

            # Sort the cube into the correct shape, so that the order of coordinates
            # is the same as in the source data
            coord_map = sorted(coord_map, key=lambda x: x[1])
            transpose_order = [coord[2] for coord in coord_map]
            cube.transpose(transpose_order)
            output.append(cube)

        log_memory_profile("GeneralGriddedCollocator Finished")

        return output

    def _set_multi_value_kernel(self, kernel_val, values, indices):
        # This kernel returns multiple values:
        for idx, val in enumerate(kernel_val):
            values[idx][indices] = val

    def _set_single_value_kernel(self, kernel_val, values, indices):
        values[0][indices] = kernel_val

    def _create_collocated_cube(self, src_data, data, coords):
        """Creates a cube using the metadata from the source cube and supplied data.

        :param src_data: ungridded data that was to be collocated
        :param data: collocated data values
        :param coords: coordinates for output cube
        :return: cube of collocated data
        """
        dim_coords_and_dims = []
        for idx, coord in enumerate(coords):
            dim_coords_and_dims.append((coord, idx))
        cube = GriddedData(data, standard_name=src_data.standard_name,
                           long_name=src_data.long_name,
                           var_name=src_data.var_name,
                           units=src_data.units,
                           dim_coords_and_dims=dim_coords_and_dims)
        return cube


class CubeCellConstraint(CellConstraint):
    #TODO Delete me
    """Constraint for constraining HyperPoints to be within an iris.coords.Cell.
    """

    def constrain_points(self, sample_point, data):
        """Returns HyperPoints lying within a cell.
        :param sample_point: HyperPoint of cells defining sample region
        :param data: list of HyperPoints to check
        :return: HyperPointList of points found within cell
        """
        con_points = HyperPointList()
        for point in data:
            include = True
            for idx in range(HyperPoint.number_standard_names):
                cell = sample_point[idx]
                if cell is not None:
                    if not (np.min(cell.bound) <= point[idx] < np.max(cell.bound)):
                        include = False
            if include:
                con_points.append(point)
        return con_points


class BinningCubeCellConstraint(IndexedConstraint):
    #TODO Delete me
    """Constraint for constraining HyperPoints to be within an iris.coords.Cell.

    Uses the index_data method to bin all the points
    """

    def __init__(self):
        super(BinningCubeCellConstraint, self).__init__()
        self.grid_cell_bin_index = None

    def constrain_points(self, sample_point, data):
        """Returns HyperPoints lying within a cell.

        This implementation returns the points that have been stored in the
        appropriate bin by the index_data method.
        :param sample_point: HyperPoint of indices of cells defining sample region
        :param data: list of HyperPoints to check
        :return: HyperPointList of points found within cell
        """
        point_list = self.grid_cell_bin_index.get_points_by_indices(sample_point)
        con_points = HyperPointList()
        if point_list is not None:
            for point in point_list:
                con_points.append(data[point])
        return con_points


class BinnedCubeCellOnlyConstraint(Constraint):
    """Constraint for constraining HyperPoints to be within an iris.coords.Cell. With an iterator which only
    travels over those cells with a value in.

    Uses the index_data method to bin all the points.
    """

    def __init__(self):
        super(BinnedCubeCellOnlyConstraint, self).__init__()
        self.grid_cell_bin_index_slices = None

    def constrain_points(self, sample_point, data):
        pass

    def get_iterator(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, output_data):

        for out_indices, slice_start_end in self.grid_cell_bin_index_slices.get_iterator():
            if not missing_data_for_missing_sample or points.data[out_indices] is not np.ma.masked:
                # iterate through the points which are within the same cell

                con_points = HyperPointList()
                slice_indicies = slice(*slice_start_end)

                for x in self.grid_cell_bin_index_slices.sort_order[slice_indicies]:
                    con_points.append(data_points[x])

                hp_values = [None] * HyperPoint.number_standard_names
                for (hpi, ci, shi) in coord_map:
                    hp_values[hpi] = coords[ci].points[out_indices[shi]]
                hp = HyperPoint(*hp_values)

                yield out_indices, hp, con_points

    def get_iterator_for_data_only(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points,
                                   values):
        """
        The method returns an iterator over the output indices and a numpy array slice of the data values. This may not
        be called by all collocators who may choose to iterate over all sample points instead.

        :param missing_data_for_missing_sample: If true anywhere there is missing data on the sample then final point is
         missing; otherwise just use the sample
        :param coord_map: Not needed for the data only kernel
        :param coords: Not needed for the data only kernel
        :param data_points: The (non-masked) data points
        :param shape: Not needed
        :param points: The original points object, these are the points to collocate
        :param values: Not needed
        :return: Iterator which iterates through (sample indices and data slice) to be placed in these points
        """
        data_points_sorted = data_points.data[self.grid_cell_bin_index_slices.sort_order]
        if missing_data_for_missing_sample:
            for out_indices, slice_start_end in self.grid_cell_bin_index_slices.get_iterator():
                # Remap the indices to match the data coordinate order, using the coord_map provided
                remapped_indices = tuple(np.asarray(out_indices)[[c[2] for c in sorted(coord_map, key=lambda x: x[1])]])
                if points.data[remapped_indices] is not np.ma.masked:
                    data_slice = data_points_sorted[slice(*slice_start_end)]
                    yield out_indices, data_slice
        else:
            for out_indices, slice_start_end in self.grid_cell_bin_index_slices.get_iterator():
                data_slice = data_points_sorted[slice(*slice_start_end)]
                yield out_indices, data_slice


def make_coord_map(points, data):
    """
    Create a map for how coordinates from the sample points map to the standard hyperpoint coordinates. Ignoring
    coordinates which are not present in the data
    :param points: sample points
    :param data: data to map
    :return: list of tuples, each tuple is index of coordinate to use
    tuple is (hyper point coord index, sample point coord index, output coord index)
    """
    # If there are coordinates in the sample grid that are not present for the data,
    # omit the from the set of coordinates in the output grid. Find a mask of coordinates
    # that are present to use when determining the output grid shape.
    coordinate_mask = [False if c is None else True for c in data.find_standard_coords()]

    # Find the mapping of standard coordinates to those in the sample points and those to be used
    # in the output data.
    return _find_standard_coords(points, coordinate_mask)


def _find_standard_coords(cube, coordinate_mask):
    """Finds the mapping of sample point coordinates to the standard ones used by HyperPoint.

    :param cube: cube among the coordinates of which to find the standard coordinates
    :param coordinate_mask: list of booleans indicating HyperPoint coordinates that are present
    :return: list of tuples relating index in HyperPoint to index in coords and in coords to be iterated over
    """
    coord_map = []
    cube_coord_lookup = {}

    cube_coords = cube.coords()
    for idx, coord in enumerate(cube_coords):
        cube_coord_lookup[coord] = idx

    shape_idx = 0
    for hpi, name in enumerate(HyperPoint.standard_names):
        if coordinate_mask[hpi]:
            # Get the dimension coordinates only - these correspond to dimensions of data array.
            cube_coords = cube.coords(standard_name=name, dim_coords=True)
            if len(cube_coords) > 1:
                msg = ('Expected to find exactly 1 coordinate, but found %d. They were: %s.'
                       % (len(cube_coords), ', '.join(coord.name() for coord in cube_coords)))
                raise cis.exceptions.CoordinateNotFoundError(msg)
            elif len(cube_coords) == 1:
                coord_map.append((hpi, cube_coord_lookup[cube_coords[0]], shape_idx))
                shape_idx += 1
    return coord_map


def _find_longitude_range(coords):
    """Finds the start of the longitude range, assumed to be either 0,360 or -180,180
    :param coords: coordinates to check
    :return: starting value for longitude range or None if no longitude coordinate found
    """
    low = None
    for coord in coords:
        if coord.standard_name == 'longitude':
            low = 0.0
            min_val = coord.points.min()
            if min_val < 0.0:
                low = -180.0
    return low


def _fix_longitude_range(coords, data_points):
    """Sets the longitude range of the data points to match that of the sample coordinates.
    :param coords: coordinates for grid on which to collocate
    :param data_points: HyperPointList or GriddedData of data to fix
    """
    range_start = _find_longitude_range(coords)
    if range_start is not None:
        data_points.set_longitude_range(range_start)
